"""
Configuration Loader
--------------------
Various functions that assist in loading initial configuration for the
control plane.

The control plane accepts a main configuration file from the environment
variable ``SOVEREIGN_CONFIG`` which follows the format:

.. code-block:: none

   <scheme>://path[,<scheme>://path,...]

Examples:

.. code-block:: none

   # Single file
     file:///etc/sovereign.yaml

   # Multiple files (comma separated)
     file:///etc/sovereign/common.yaml,file:///etc/sovereign/dev.yaml

   # Other types of sources
     http://config.myserver.com/environments/dev.yaml

   # Other types of formats
     http+json://config.myserver.com/environments/dev.json
     http+jinja://config.myserver.com/environments/dev.j2
     http+yaml://config.myserver.com/environments/dev.yaml

"""
import os
import json
import importlib
import yaml
import jinja2
import requests
from pkg_resources import resource_string


serializers = {
    'yaml': yaml.safe_load,
    'json': json.loads,
    'jinja': jinja2.Template
}


def load_file(path, loader):
    with open(path) as f:
        contents = f.read()
        return serializers[loader](contents)


def load_package_data(path, loader):
    pkg, pkg_file = path.split(':')
    data = resource_string(pkg, pkg_file)
    try:
        data = data.decode()
    except AttributeError:
        pass
    return serializers[loader](data)


def load_http(path, loader):
    data = requests.get(path).text
    return serializers[loader](data)


def load_env(variable, loader=None):
    data = os.getenv(variable)
    if loader is not None and data is not None:
        return serializers[loader](data)
    return data


def load_module(name, _=None):
    return importlib.import_module(name)


loaders = {
    'file': load_file,
    'pkgdata': load_package_data,
    'http': load_http,
    'https': load_http,
    'env': load_env,
    'module': load_module
}


def load(spec):
    if '://' not in spec:
        return spec

    serialization = 'yaml'
    scheme, path = spec.split('://')
    if '+' in scheme:
        scheme, serialization = scheme.split('+')
    if 'http' in scheme:
        path = '://'.join([scheme, path])
    return loaders[scheme](path, serialization)



import os
import time
import traceback
from datetime import datetime, timedelta
from quart import (
    Quart,
    g,
    request,
    jsonify,
    redirect,
    url_for,
    make_response
)
from flask_log_request_id import RequestID, current_request_id
try:
    import sentry_sdk
    from sentry_asgi import SentryMiddleware
except ImportError:
    sentry_sdk = None

from sovereign import statsd, ENVIRONMENT, SENTRY_DSN
from sovereign.logs import LOG
from sovereign.views import (
    crypto,
    discovery,
    healthchecks,
    admin,
)


def init_app():
    application: Quart = Quart(__name__)
    RequestID(application)

    application.config['JSONIFY_PRETTYPRINT_REGULAR'] = True
    application.config['RESPONSE_TIMEOUT'] = 5
    application.config['BODY_TIMEOUT'] = 5
    application.config['MAX_CONTENT_LENGTH'] = 1024 * 4
    application.host = os.getenv('SOVEREIGN_HOST', '0.0.0.0')
    application.port = int(os.getenv('SOVEREIGN_PORT', '8080'))

    application.register_blueprint(admin.blueprint)
    application.register_blueprint(discovery.blueprint)
    application.register_blueprint(healthchecks.blueprint)
    application.register_blueprint(crypto.blueprint)

    for handler in application.logger.handlers:
        application.logger.removeHandler(handler)

    # pylint: disable=unused-variable

    @application.route('/')
    def index():
        return redirect(url_for('admin.display_config'))

    @application.route('/favicon.ico')
    async def favicon_stub():
        expiry_time = datetime.utcnow() + timedelta(weeks=4)
        body, status, headers = (
            '', 200,
            {'Expires': expiry_time.strftime("%a, %d %b %Y %H:%M:%S GMT")}
        )
        response = await make_response(body, status, headers)
        return response

    @application.before_request
    def time_request():
        g.request_start_time = time.time()
        g.request_time = lambda: (time.time() - g.request_start_time) * 1000  # Milliseconds

    @application.before_request
    def add_logger():
        g.log = LOG.bind(
            uri_path=request.path,
            uri_query=request.query_string or '-',
            src_ip=request.remote_addr,
            site=request.headers.get('host', '-'),
            method=request.method,
            user_agent=request.headers.get('user-agent', '-'),
            request_id=current_request_id(),
            env=ENVIRONMENT,
            worker_pid=os.getpid()
        )

    @application.errorhandler(Exception)
    def exception_handler(e: Exception):
        error = {
            'error': repr(e),
            'request_id': current_request_id(),
            'traceback': traceback.format_exc()
        }
        g.log = g.log.bind(**error)
        return jsonify(error), getattr(e, 'status', 500)

    @application.after_request
    def log_request(response):
        duration = g.request_time()
        g.log.msg(
            code=response.status_code,
            duration=duration
        )
        if 'discovery' in str(request.endpoint):
            tags = [
                f'path:{request.path}',
                f'code:{response.status_code}',
            ]
            statsd.timing('rq_ms', value=duration, tags=tags)
        return response

    if SENTRY_DSN and sentry_sdk:
        sentry_sdk.init(SENTRY_DSN)
        application = SentryMiddleware(application)

    return application


app = init_app()


if __name__ == '__main__':
    app.run(host='localhost', port=8080)



from __future__ import division

def round_to_100(sequence):
    """ Resolves issue when dividing 100 results in a set of numbers that don't add up to 100 again

        E.g. 100 / 3 = 33
             33 * 3 = 99

        For the above example, this function returns [33, 33, 34]
     """
    if sum(sequence) != 100:
        sequence[-1] = 100 - sum(sequence[:-1])
    return sequence


def fit_weights(clusters):
    weights = [cluster['weight'] for cluster in clusters]
    for cluster, newly_assigned_weight in zip(clusters, round_to_100(weights)):
        cluster['weight'] = newly_assigned_weight
    return clusters


def normalize_weights(sequence):
    total = sum(sequence)
    for item in sequence:
        yield int((item / total) * 100)



from socket import gethostbyname_ex
from socket import gaierror as dns_error
from werkzeug.exceptions import Gone
from sovereign import CONFIG, statsd
from sovereign.decorators import memoize


@memoize(5)
def resolve(address):
    try:
        with statsd.timed('dns.resolve_ms', tags=[f'address:{address}'], use_ms=True):
            _, _, addresses = gethostbyname_ex(address)
    except dns_error:
        raise Gone(f'Failed to resolve DNS hostname: {address}')
    else:
        return addresses


def healthchecks_enabled(healthchecks):
    for healthcheck in healthchecks:
        if healthcheck.get('path') in ('no', False):
            return False
    return True


def upstream_requires_tls(cluster):
    for host in cluster.get('hosts', []):
        if '443' in str(host.get('port')):
            return True
    return False


def list_regions():
    return CONFIG.get('regions', [])


def remove_tls_certificates(listener):
    for filter_chain in listener['filter_chains']:
        try:
            # Traverse down, each line separate for easier debugging
            tls_context = filter_chain['tls_context']
            common_tls_context = tls_context['common_tls_context']
            tls_certificates = common_tls_context['tls_certificates']
        except KeyError:
            pass
        else:
            for certificate in tls_certificates:
                certificate['private_key'] = {'inline_string': '<REDACTED>'}



import random
from quart import Blueprint
from sovereign import XDS_TEMPLATES
from sovereign import discovery
from sovereign.sources import load_sources
from sovereign.utils.mock import mock_discovery_request


blueprint = Blueprint('healthchecks', __name__)

xds_version = list(XDS_TEMPLATES.keys())[-1]


@blueprint.route('/healthcheck')
def health_check():
    return 'OK'


@blueprint.route('/deepcheck')
def deep_check():
    template = random.choice(
        list(XDS_TEMPLATES[xds_version].keys())
    )
    discovery.response(
        mock_discovery_request(),
        xds=template,
        version=xds_version,
        debug=True
    )
    load_sources(service_cluster='', debug=True)
    return 'OK'



import jsonschema
from quart import Blueprint, request, jsonify, g
from flask_log_request_id import current_request_id
from sovereign.utils.crypto import encrypt, decrypt, generate_key

blueprint = Blueprint('crypto', __name__)

schema = {
    'title': 'Encryption parameter schema',
    "type": "object",
    'required': [
        'data'
    ],
    "properties": {
        'data': {
            'description': 'Data to be encrypted/decrypted.',
            'type': 'string',
            'minLength': 1,
            'maxLength': 65535
        },
        'key': {
            'description': 'Secret key to use when encrypting/decrypting. '
                           'Default is the applications own key, unless decrypting.',
            'type': 'string',
            'minLength': 44,
            'maxLength': 44
        }
    }
}


@blueprint.route('/crypto')
@blueprint.route('/crypto/help')
def _help():
    return jsonify(schema)


@blueprint.route('/crypto/decrypt', methods=['POST'])
async def _decrypt():
    req = await request.get_json(force=True)
    jsonschema.validate(req, schema)
    try:
        ret = {
            'result': decrypt(req['data'], req['key'])
        }
        code = 200
    except KeyError:
        ret = {
            'error': 'A key must be supplied to use for decryption'
        }
        code = 400
    return jsonify(ret), code


@blueprint.route('/crypto/encrypt', methods=['POST'])
async def _encrypt():
    req = await request.get_json(force=True)
    jsonschema.validate(req, schema)
    ret = {
        'result': encrypt(**req)
    }
    return jsonify(ret)


@blueprint.route('/crypto/decryptable', methods=['POST'])
async def _decryptable():
    req = await request.get_json(force=True)
    jsonschema.validate(req, schema)
    try:
        decrypt(req['data'])
        ret = {}
        code = 200
    except KeyError as e:
        ret = {
            'error': str(e),
            'request_id': current_request_id()
        }
        code = 500
    return jsonify(ret), code


@blueprint.route('/crypto/generate_key')
def _generate_key():
    ret = {
        'result': generate_key()
    }
    return jsonify(ret)


@blueprint.errorhandler(jsonschema.SchemaError)
def schema_handler(exception):
    error = {
        'error': f'There was a problem with the server\'s schema.',
        'request_id': current_request_id()
    }
    g.log = g.log.bind(**error, exception=repr(exception))
    return jsonify(error), 500


@blueprint.errorhandler(jsonschema.ValidationError)
def validation_handler(e: jsonschema.ValidationError):
    error = {
        'error': f'{e.__class__.__name__}: {e.message}',
        'request_id': current_request_id()
    }
    g.log = g.log.bind(**error)
    return jsonify(error), 400



"""
Discovery
---------

Functions used to render and return discovery responses to Envoy proxies.

The templates are configurable. `todo See ref:Configuration#Templates`
"""
import yaml
from yaml.parser import ParserError
from sovereign import XDS_TEMPLATES, TEMPLATE_CONTEXT, DEBUG, statsd
from sovereign.decorators import envoy_authorization_required
from sovereign.sources import load_sources

try:
    default_templates = XDS_TEMPLATES['default']
except KeyError:
    raise KeyError(
        'Your configuration should contain default templates. For more details, see '
        'https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/tutorial.html#create-templates '
    )


@statsd.timed('discovery.version_hash_ms', use_ms=True)
def version_hash(*args) -> str:
    """
    Creates a 'version hash' to be used in envoy Discovery Responses.
    """
    return str(hash(repr(args)))


@envoy_authorization_required
def response(request, xds, version, debug=DEBUG) -> dict:
    """
    A Discovery **Request** typically looks something like:

    .. code-block:: json

        {
            "version_info": "0",
            "node": {
                "cluster": "T1",
                "build_version": "<revision hash>/<version>/Clean/RELEASE",
                "metadata": {
                    "auth": "..."
                }
            }
        }

    When we receive this, we give the client the latest configuration via a
    Discovery **Response** that looks something like this:

    .. code-block:: json

        {
            "version_info": "abcdef1234567890",
            "resources": []
        }

    The version_info is derived from :func:`sovereign.discovery.version_hash`

    :param request: An envoy Discovery Request
    :param xds: what type of XDS template to use when rendering
    :param version: what template version to render for (i.e. envoy 1.7.0, 1.8.0)
    :param debug: switch to control instance loading / exception raising
    :return: An envoy Discovery Response
    """
    cluster = request['node']['cluster']
    metrics_tags = [
        f'xds_type:{xds}',
        f'partition:{cluster}'
    ]
    with statsd.timed('discovery.total_ms', use_ms=True, tags=metrics_tags):
        context = {
            'instances': load_sources(cluster, debug=debug),
            'resource_names': request.get('resource_names', []),
            'debug': debug,
            **TEMPLATE_CONTEXT
        }
        template = XDS_TEMPLATES.get(version, default_templates)[xds]
        config_version = version_hash(context, template, request['node'])
        if config_version == request.get('version_info', '0'):
            return {'version_info': config_version}

        with statsd.timed('discovery.render_ms', use_ms=True, tags=metrics_tags):
            rendered = template.render(discovery_request=request, **context)
        try:
            configuration = yaml.load(rendered)
            configuration['version_info'] = config_version
            return configuration
        except ParserError:
            if debug:
                raise
            raise ParserError('Failed to render configuration')



import os
from setuptools import setup, find_packages

project_name = 'sovereign'

with open('VERSION') as v:
    project_version = v.read().strip()


def requirements():
    try:
        with open('requirements.txt') as f:
            return f.read().splitlines()
    except FileNotFoundError:
        print(os.path.abspath(os.curdir))
        print(os.listdir('.'))
        raise


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name=project_name,
    version=project_version,
    python_requires='>=3.7.0',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    zip_safe=False,
    include_package_data=True,
    url='https://bitbucket.org/atlassian/sovereign',
    license='Apache-2.0',
    author='Vassilios Syrakis',
    author_email='vsyrakis@atlassian.com',
    description='Envoy Proxy control-plane written in Python',
    long_description=readme(),
    long_description_content_type='text/markdown',
    entry_points={
        "sovereign.sources": [
            "service_broker = sovereign.sources.service_broker:ServiceBroker",
            "file = sovereign.sources.file:File",
            "inline = sovereign.sources.inline:Inline",
        ]
    },
    scripts=[
        'bin/sovereign'
    ],
    install_requires=requirements(),
    command_options={
        'build_sphinx': {
            'project': ('setup.py', project_name),
            'version': ('setup.py', project_version),
            'release': ('setup.py', project_version),
            'source_dir': ('setup.py', 'docs'),
            'config_dir': ('setup.py', 'docs'),
            'build_dir': ('setup.py', 'src/sovereign/docs'),
            'builder': ('setup.py', 'html')
        }
    },
    extras_require={
        'sentry': ['sentry-sdk', 'sentry-asgi']
    }
)
